# -*- coding: utf-8 -*-
"""
Created on Thu Sep  5 13:05:12 2019

@author: icruicks
"""

import pandas as pd, GraphLearning as gl, networkx as nx, numpy as np
from scipy.sparse import csr_matrix, diags
from scipy.sparse.linalg import eigs
from scipy.stats import mode
import os, warnings
from matplotlib import pyplot as plt

from sklearn.metrics import adjusted_rand_score, adjusted_mutual_info_score, confusion_matrix, silhouette_score
from sklearn.decomposition import LatentDirichletAllocation
from sklearn.neighbors import kneighbors_graph
from sklearn.cluster import SpectralClustering, AgglomerativeClustering, KMeans
from sklearn.cluster import SpectralCoclustering
from sknetwork.clustering import BiLouvain, Louvain, modularity
from Leiden_Clustering import Leiden_Unimodal
from sklearn.preprocessing import normalize

def create_twitter_network(users, edgelist):
    n = len(users)
    network = np.zeros([n,n])

    for i in range(n):
        pers_i = users[i]
        res_1 = edgelist[np.where(edgelist[:,0] ==pers_i)]
        for j in range(n):
            pers_j = users[j]
            res_2 = res_1[np.where(res_1[:,1] ==pers_j)]
            try:
                network[i,j] = res_2[:,2]
            except:
                network[i,j] = 0

    return csr_matrix(network)

def create_twitter_attribute(users, content):
    n = len(users)
    attributes = np.unique(content[:,0])
    m = len(attributes)
    network = np.zeros([n,m])
    for i in range(n):
        pers_i = users[i]
        res_1 = content[np.where(content[:,1] ==pers_i)]
        for j in range(m):
            attr_j = attributes[j]
            res_2 = res_1[np.where(res_1[:,0] == attr_j)]
            try:
                network[i,j] = res_2[:,2]
            except:
                network[i,j] = 0
    return csr_matrix(network)


def print_results(y_preds, y, resolutions, weights, modal_modularities, iterations, best_iterations, model_name):
    aris = [adjusted_rand_score(y, y_pred) for y_pred in y_preds]
    amis = [adjusted_mutual_info_score(y, y_pred) for y_pred in y_preds]
    resolutions = np.array(resolutions)
    weights = np.array(weights)
    iterations =np.array(iterations)
    modal_modularities = np.array(modal_modularities)
    print("\n")
    print(model_name+" Results:")
    print("Adjusted Rand Score: {} +/- {}".format(np.mean(aris), np.std(aris)))
    print("Adjusted Mutual Information: {} +/- {}".format(np.mean(amis), np.std(amis)))
    # print("Confusion Matrix of best ARI result: \n {}".format(confusion_matrix(y, y_preds[np.argmax(aris)])))
    print("Avg. Number of clusters found: {}".format(np.mean([len(np.unique(y_pred)) for y_pred in y_preds])))
    print("optimal modal resolutions: {} +/- {}".format(np.mean(resolutions, axis=0), np.std(resolutions, axis=0)))
    print("optimal modal weights: {} +/- {}".format(np.mean(weights, axis=0), np.std(weights, axis=0)))
    print("Number of iterations used: {} +/- {}".format(np.mean(iterations), np.std(iterations)))
    print("Best result found at iteration: {} +/- {}".format(np.mean(best_iterations), np.std(best_iterations)))
    print("optimal weighted modularity: {} +/- {}".format(np.mean(modal_modularities, axis=0), np.std(modal_modularities, axis=0)))

def read_in_data(dataset):
    ''' 
    Read in Data
    '''
    if dataset =='cora':
        content = pd.read_csv(os.path.join('Data','Cora','cora_content.csv'), header=0, index_col=0)
        links = pd.read_csv(os.path.join('Data','Cora','cora_network_links.csv'), header=0, index_col=None)
        network = nx.from_pandas_edgelist(links, source='Source', target='Target', create_using=nx.DiGraph())
        df = nx.to_pandas_adjacency(network)
        content = content.reindex(labels=df.index, axis='index')
        adjacency = df.values
        X = content.iloc[:,:-1]
        y = content.iloc[:,-1].astype('category').cat.codes
        nearest_neighbors = 40
    elif dataset=='citeseer':
        content = pd.read_csv(os.path.join('Data','CiteSeer','citeseer_content.csv'), header=0, index_col=0, dtype={0: object})
        links = pd.read_csv(os.path.join('Data','CiteSeer','citeseer_network_links.csv'), header=0, index_col=None, dtype=object)
        discrepancies = list(np.setdiff1d(links['Source'], content.index)) + list(np.setdiff1d(links['Target'], content.index))
        links = links[~links['Source'].isin(discrepancies)]
        links = links[~links['Target'].isin(discrepancies)]
        content = content[~content.index.isin(discrepancies)]
        network = nx.from_pandas_edgelist(links, source='Source', target='Target', create_using=nx.DiGraph())
        df = nx.to_pandas_adjacency(network)
        content = content.reindex(labels=df.index, axis='index')
        adjacency = df.values
        X = content.iloc[:,:-1]
        y = content.iloc[:,-1].astype('category').cat.codes
        nearest_neighbors = 40
    elif dataset =='flickr':
        content = pd.read_csv(os.path.join('Data','Flickr','attributes.csv'), header=None)
        adjacency = pd.read_csv(os.path.join('Data','Flickr','network.csv'), header= None).to_numpy()
        y = pd.read_csv(os.path.join('Data','Flickr','labels.csv'), header = None)[0]
        X = content.to_numpy()
        nearest_neighbors = 60
    elif dataset =='blogcatalog':
        content = pd.read_csv(os.path.join('Data','BlogCatalog','attributes.csv'), header=None)
        adjacency = pd.read_csv(os.path.join('Data','BlogCatalog','network.csv'), header= None).to_numpy()
        y = pd.read_csv(os.path.join('Data','BlogCatalog','labels.csv'), header = None)[0]
        X = content.to_numpy()
        nearest_neighbors = 60
    elif dataset =='3sources':
        from scipy.io import mmread
        from functools import reduce
        bbc_content = csr_matrix(mmread(os.path.join('Data','3Sources','3sources_bbc.mtx'))).T
        guardian_content = csr_matrix(mmread(os.path.join('Data','3Sources','3sources_guardian.mtx'))).T
        reuters_content = csr_matrix(mmread(os.path.join('Data','3Sources','3sources_reuters.mtx'))).T
        labels = pd.read_csv(os.path.join('Data','3Sources','labels.csv'), header = None)[1].astype('category').cat.codes
        reuters_idx = np.ravel(pd.read_csv(os.path.join("Data","3sources","3sources_reuters.docs"), header=None).values)-1
        bbc_idx = np.ravel(pd.read_csv(os.path.join("Data","3sources","3sources_bbc.docs"), header=None).values)-1
        guardian_idx = np.ravel(pd.read_csv(os.path.join("Data","3sources","3sources_guardian.docs"), header=None).values)-1
        master_idx = reduce(np.intersect1d, (bbc_idx, guardian_idx, reuters_idx))
        bbc_X = bbc_content[np.where(np.isin(bbc_idx, master_idx))]
        reuters_X = reuters_content[np.where(np.isin(reuters_idx, master_idx))]
        guardian_X = guardian_content[np.where(np.isin(guardian_idx, master_idx))]
        y = labels[master_idx]
        nearest_neighbors = 20
    elif dataset =='wiki':
        content = np.loadtxt(os.path.join('Data','Wiki','tfidf.txt'))
        links = np.loadtxt(os.path.join('Data','Wiki','graph.txt'))
        labels = np.loadtxt(os.path.join('Data','Wiki','group.txt'))
        X = pd.crosstab(content[:,0], content[:,1], values=content[:,2], aggfunc=sum)
        X = X.reindex(index=labels[:,0])
        X = np.nan_to_num(X.values)
        adjacency = pd.crosstab(links[:,0], links[:,1])
        adjacency = adjacency.reindex(index=labels[:,0], columns=labels[:,0])
        adjacency = csr_matrix(np.nan_to_num(adjacency.values))
        y= labels[:,1]
        nearest_neighbors = int(np.ceil(np.sqrt(labels.shape[0])))
    elif dataset in ['football', 'olympics', 'politicsie', 'politicsuk', 'rugby']:
        user_ids = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+".ids"))
        content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-follows.mtx"))[1:,:]
        follows_network = csr_matrix(pd.crosstab(content[:,0], content[:,1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        # follows_network = create_twitter_network(user_ids, content)
        content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-mentions.mtx"))[1:,:]
        mentions_network = csr_matrix(pd.crosstab(content[:,0], content[:,1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        # mentions_network = create_twitter_network(user_ids, content)
        content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-retweets.mtx"))[1:,:]
        retweets_network = csr_matrix(pd.crosstab(content[:,0], content[:,1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        # retweets_network = create_twitter_network(user_ids, content)
        content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-lists500.mtx"))[1:,:]
        list_content = csr_matrix(pd.crosstab(content[:,1], content[:,0]).reindex(index=user_ids, fill_value=0).values)
        # list_content = create_twitter_attribute(user_ids, content)
        content = np.loadtxt(os.path.join("Data","Twitter",dataset,dataset+"-tweets500.mtx"))[1:,:]
        text_content = csr_matrix(pd.crosstab(content[:,1], content[:,0]).reindex(index=user_ids, fill_value=0).values)
        # text_content = create_twitter_attribute(user_ids, content)
        y = pd.read_csv(os.path.join("Data","Twitter",dataset,"labels.csv"), header = None)[1].astype('category').cat.codes
        nearest_neighbors = int(np.ceil(np.sqrt(user_ids.shape[0])))
    elif dataset == 'fftwyt':
        links = pd.read_csv(os.path.join('Data','ff-tw-yt','raw_data.csv'), header= None)
        links_ff = links.loc[links[2] == 'ff']
        links_tw = links.loc[links[2] == 'tw']
        links_yt = links.loc[links[2] == 'yt']
        user_ids = pd.unique(links[[0, 1]].values.ravel('K'))
        ff_network = csr_matrix(pd.crosstab(links_ff[0], links_ff[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        tw_network = csr_matrix(pd.crosstab(links_tw[0], links_tw[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        yt_network = csr_matrix(pd.crosstab(links_yt[0], links_yt[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    elif dataset =='aucs':
        links = pd.read_csv(os.path.join('Data','AUCS','links.csv'), header= None)
        labels = pd.read_csv(os.path.join('Data','AUCS','attributes.csv'), header= None).fillna("none")
        user_ids= labels[0].values
        y= labels[1].astype('category').cat.codes
        links_lunch = links.loc[links[2] == 'lunch']
        lunch_network = csr_matrix(pd.crosstab(links_lunch[0], links_lunch[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        links_facebook = links.loc[links[2] == 'facebook']
        facebook_network = csr_matrix(pd.crosstab(links_facebook[0], links_facebook[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        links_coauthor = links.loc[links[2] == 'coauthor']
        coauthor_network = csr_matrix(pd.crosstab(links_coauthor[0], links_coauthor[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        links_work = links.loc[links[2] == 'work']
        work_network = csr_matrix(pd.crosstab(links_work[0], links_work[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
        links_leisure = links.loc[links[2] == 'leisure']
        leisure_network = csr_matrix(pd.crosstab(links_leisure[0], links_leisure[1]).reindex(index=user_ids, columns=user_ids, fill_value=0).values)
    else:
        raise ValueError(dataset+" not found")
    
    ''' 
    Create the base local graphs and clusters 
    '''
    kNNs =[]
    clusters =[]
        
    if dataset in ['cora', 'citeseer', 'blogcatalog', 'flickr', 'wiki']:
        # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=10)]
        # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
        cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
        parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
        modularities = [modularity(adjacency, part) for part in parts]
        # print("Modularity of network clustering: {}".format(np.max(modularities)))
        defacto_network_subgroups = parts[np.argmax(modularities)]
        clusters.append(defacto_network_subgroups)
        # kNN = np.maximum(adjacency, adjacency.T)
        # kNNs.append(csr_matrix(kNN))
        kNNs.append(csr_matrix(adjacency))
        
        kNN = kneighbors_graph(X, metric='cosine', mode='connectivity', n_neighbors=nearest_neighbors)
        kNN = kNN.minimum(kNN.T)
        kNNs.append(kNN)
        if dataset in ['cora', 'citeseer', 'blogcatalog', 'wiki']:
            # clstr = SpectralClustering(n_clusters=len(np.unique(y)), affinity='rbf', n_jobs=-1)
            # clstr = AgglomerativeClustering(n_clusters=len(np.unique(y)))
            # clstr = KMeans(n_clusters=len(np.unique(y)), random_state=1234)
            # clstr = OPTICS(metric='cosine')
            # content_subgroups = clstr.fit_predict(X)
            lda = LatentDirichletAllocation(n_components=len(np.unique(y)), learning_decay= 0.5, random_state=1234)
            lda_probs = lda.fit_transform(X)
            _,content_subgroups = np.unique(np.argmax(lda_probs, axis=1),return_inverse=True)
        elif dataset in ['flickr']:
            attrib_cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
            attrib_parts = [clstr.fit_transform(kNN) for clstr in cluster_iterations]
            attrib_modularities = [modularity(kNN, part) for part in attrib_parts]
            content_subgroups = attrib_parts[np.argmax(attrib_modularities)]
        # print("Silhouette of attributes clustering: {}".format(silhouette_score(X, content_subgroups)))
        clusters.append(content_subgroups)

    elif dataset == '3sources':
        for X in [bbc_X, reuters_X, guardian_X]:
            clstr = AgglomerativeClustering(n_clusters=len(np.unique(y)))
            content_subgroups = clstr.fit_predict(X.toarray())
            # print("Silhouette of text clustering: {}".format(silhouette_score(X, content_subgroups)))
            clusters.append(content_subgroups)
            kNN = kneighbors_graph(X, metric='cosine', mode='connectivity', n_neighbors=nearest_neighbors)
            kNN = kNN.minimum(kNN.T)
            kNNs.append(kNN)
    elif dataset in ['football', 'olympics', 'politicsie', 'politicsuk', 'rugby']:
        for X in [list_content, text_content]:
            lda = LatentDirichletAllocation(n_components=len(np.unique(y)), random_state=1234)
            lda_probs = lda.fit_transform(X)
            _,content_subgroups = np.unique(np.argmax(lda_probs, axis=1),return_inverse=True)
            # print("Silhouette of attributes clustering: {}".format(silhouette_score(X, content_subgroups)))
            clusters.append(content_subgroups)
            kNN = kneighbors_graph(X, metric='cosine', mode='connectivity', n_neighbors=nearest_neighbors)
            kNN = kNN.minimum(kNN.T)
            kNNs.append(kNN)
        for adjacency in [follows_network, mentions_network, retweets_network]:
            # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
            # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
            cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
            parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
            modularities = [modularity(adjacency, part) for part in parts]
            # print("Modularity of network clustering: {}".format(np.max(modularities)))
            defacto_network_subgroups = parts[np.argmax(modularities)]
            clusters.append(defacto_network_subgroups)
            # kNN = np.maximum(adjacency, adjacency.T)
            kNNs.append(csr_matrix(adjacency))
    elif dataset == 'fftwyt':
        for adjacency in [ff_network, tw_network, yt_network]:
            # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
            # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
            cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
            parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
            modularities = [modularity(adjacency, part) for part in parts]
            # print("Modularity of network clustering: {}".format(np.max(modularities)))
            defacto_network_subgroups = parts[np.argmax(modularities)]
            clusters.append(defacto_network_subgroups)
            # kNN = np.maximum(adjacency, adjacency.T)
            kNNs.append(csr_matrix(adjacency))
    elif dataset == 'aucs':
        for adjacency in [lunch_network, facebook_network, coauthor_network, work_network, leisure_network]:
            # cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=5)]
            # parts = [clstr.fit(adjacency).labels_ for clstr in cluster_iterations]
            cluster_iterations = [Leiden_Unimodal(resolution=i) for i in np.linspace(0.1,2.0,10)]
            parts = [clstr.fit_transform(adjacency) for clstr in cluster_iterations]
            modularities = [modularity(adjacency, part) for part in parts]
            # print("Modularity of network clustering: {}".format(np.max(modularities)))
            defacto_network_subgroups = parts[np.argmax(modularities)]
            clusters.append(defacto_network_subgroups)
            # kNN = np.maximum(adjacency, adjacency.T)
            kNNs.append(csr_matrix(adjacency))
    else:
        raise ValueError(dataset+" not found")
        
    if dataset in ['cora', 'citeseer', 'blogcatalog', 'flickr', 'wiki']:
        modes = ['network', 'content']
    elif dataset == '3sources':
        modes = ['BBC','Reuters','Guardian']
    elif dataset in ['football', 'olympics', 'politicsie', 'politicsuk', 'rugby']:
        modes = ['list content', 'text content', 'follows network','mentions network','retweets network']
    elif dataset == 'aucs':
        modes = ['lunch network', 'facebook network', 'coauthor network', 'work network', 'leisure network']

    return kNNs, clusters, y, modes



for dataset in ['cora', 'citeseer', 'flickr', 'blogcatalog', 'wiki',
                '3sources', 'aucs', 'football', 'olympics', 'politicsie',
                'politicsuk', 'rugby']:
    
    print('\n')
    print('-------- '+dataset+' ----------')
    num_iterations = 20
    warnings.filterwarnings('ignore')
    
    kNNs, clusters, y, modes = read_in_data(dataset)
    
    '''
    Auto-weighted Modularity Clustering
    '''
    from Leiden_Clustering import MVMC
    
    y_preds =[]
    resolutions =[]
    weights =[]
    modularities =[]
    iterations_needed =[]
    best_iterations =[]
    for iteration in range(num_iterations):
        mvmc_clstr= MVMC()
        y_preds.append(mvmc_clstr.fit_transform(kNNs))
        resolutions.append(mvmc_clstr.best_resolutions)
        weights.append(mvmc_clstr.best_weights)
        modularities.append(mvmc_clstr.best_modularity)
        iterations_needed.append(mvmc_clstr.final_iteration)
        best_iterations.append(mvmc_clstr.best_iteration)
    print_results(y_preds, y, resolutions, weights, modularities,
                  iterations_needed, best_iterations, "Modularity Clustering for Multi-View Data")

    '''
    Grid search resolutions CPM Modualrity Clsutering
    '''
    # from Leiden_Clustering import MVDMC
    
    # y_preds =[]
    # resolutions =[]
    # weights =[]
    # modularities =[]
    # iterations_needed =[]
    # best_iterations =[]
    # for iteration in range(num_iterations):
    #     mvdmc_clstr= MVDMC(n_iterations=5)
    #     y_preds.append(mvdmc_clstr.fit_transform(kNNs))
    #     resolutions.append(mvdmc_clstr.best_resolution)
    #     weights.append(mvdmc_clstr.best_weights)
    #     modularities.append(mvdmc_clstr.best_modularity)
    #     iterations_needed.append(mvdmc_clstr.final_iteration)
    #     best_iterations.append(mvdmc_clstr.best_iteration)
    # print_results(y_preds, y, resolutions, weights, modularities,
    #               iterations_needed, best_iterations, "Modularity Clustering for Multi-View Data")
    # print("Mode of optimal modal resolutions: {}".format(mode(resolutions, axis=0)))
    
    
    
    
    
    
    
